import pyloadlvar
import string
 
#
# Helper functions to wrap pointer into python objects allows
# us to pass arbitrary pointers between the python objects
#

cdef extern from "Python.h":
	ctypedef void *PyObject
	object PyCObject_FromVoidPtr(void* cobj, void (*destr)(void *))
	void* PyCObject_AsVoidPtr(object)

cdef ptr_wrapper(void *ptr):
	return PyCObject_FromVoidPtr(ptr, NULL)

cdef void *ptr_unwrapper(object obj):
	return PyCObject_AsVoidPtr(obj)

#
# Memory management
#

cdef extern from "stdlib.h":
	ctypedef long long size_t	
	void free(void* ptr)
	void* malloc(size_t size)

cdef extern from "time.h":
	ctypedef int time_t
	
cdef char** pyStringSeqToStringArray(seq):
	
	cdef char **msgArray

	msgArray = NULL
	i = 0
	length = len(seq)

	if length != 0:
		msgArray = <char**> malloc((length+1)*sizeof(char*))
		for line in seq:
			msgArray[i] = line
			i = i + 1

		msgArray[i] = NULL
	
	return msgArray
    
#
# The all important LoadLeveler definitions
#

cdef extern from "llapi.h":

	ctypedef void LL_element
    
	enum QueryType:
		JOBS
		MACHINES
		PERF
		CLUSTERS
		WLMSTAT 
		MATRIX
		CLASSES

	enum QueryFlags:
		QUERY_ALL = (1<<0)
		QUERY_JOBID = (1<<1)
		QUERY_STEPID = (1<<2)
		QUERY_USER = (1<<3)
		QUERY_GROUP = (1<<4)
		QUERY_CLASS = (1<<5)
		QUERY_HOST = (1<<6)
		QUERY_PERF = (1<<7)
		QUERY_STARTDATE = (1<<8)
		QUERY_ENDDATE = (1<<9)
		QUERY_PROCID = (1<<10)

	enum DataFilter:
		ALL_DATA
		STATUS_LINE
		Q_LINE

	enum LL_Daemon:
		LL_STARTD
		LL_SCHEDD 
		LL_CM
		LL_MASTER
		LL_STARTER
		LL_HISTORY_FILE

	enum LLAPI_Specification:
		LL_JobManagementInteractiveClass=0
		LL_JobManagementListenSocket
		LL_JobManagementAccountNo
		LL_JobManagementSessionType
		LL_JobManagementPrinterFILE
		LL_JobManagementRestorePrinter
		LL_JobGetFirstStep=200
		LL_JobGetNextStep
		LL_JobCredential
		LL_JobName
		LL_JobStepCount
		LL_JobStepType
		LL_JobSubmitHost
		LL_JobSubmitTime
		LL_JobVersionNum
		LL_StepNodeCount=400
		LL_StepGetFirstNode
		LL_StepGetNextNode
		LL_StepMachineCount
		LL_StepGetFirstMachine
		LL_StepGetNextMachine
		LL_StepGetFirstSwitchTable
		LL_StepGetNextSwitchTable
		LL_StepGetMasterTask
		LL_StepTaskInstanceCount
		LL_StepAccountNumber
		LL_StepAdapterUsage
		LL_StepComment
		LL_StepCompletionCode
		LL_StepCompletionDate
		LL_StepEnvironment
		LL_StepErrorFile
		LL_StepExecSize
		LL_StepHostName
		LL_StepID
		LL_StepInputFile
		LL_StepImageSize
		LL_StepImmediate
		LL_StepIwd
		LL_StepJobClass
		LL_StepMessages
		LL_StepName
		LL_StepNodeUsage
		LL_StepOutputFile
		LL_StepParallelMode
		LL_StepPriority
		LL_StepShell
		LL_StepStartDate
		LL_StepDispatchTime
		LL_StepState
		LL_StepStartCount
		LL_StepCpuLimitHard
		LL_StepCpuLimitSoft
		LL_StepCpuStepLimitHard
		LL_StepCpuStepLimitSoft
		LL_StepCoreLimitHard
		LL_StepCoreLimitSoft
		LL_StepDataLimitHard
		LL_StepDataLimitSoft
		LL_StepFileLimitHard
		LL_StepFileLimitSoft
		LL_StepRssLimitHard
		LL_StepRssLimitSoft
		LL_StepStackLimitHard
		LL_StepStackLimitSoft
		LL_StepWallClockLimitHard
		LL_StepWallClockLimitSoft
		LL_StepHostList
		LL_StepHoldType
		LL_StepLoadLevelerGroup
		LL_StepGetFirstAdapterReq
		LL_StepGetNextAdapterReq
		LL_StepRestart
		LL_StepBlocking
		LL_StepTaskGeometry
		LL_StepTotalTasksRequested
		LL_StepTasksPerNodeRequested
		LL_StepTotalNodesRequested
		LL_StepSystemPriority
		LL_StepClassSystemPriority
		LL_StepGroupSystemPriority
		LL_StepUserSystemPriority
		LL_StepQueueSystemPriority
		LL_StepExecutionFactor
		LL_StepImageSize64
		LL_StepCpuLimitHard64
		LL_StepCpuLimitSoft64
		LL_StepCpuStepLimitHard64
		LL_StepCpuStepLimitSoft64
		LL_StepCoreLimitHard64
		LL_StepCoreLimitSoft64
		LL_StepDataLimitHard64
		LL_StepDataLimitSoft64
		LL_StepFileLimitHard64
		LL_StepFileLimitSoft64
		LL_StepRssLimitHard64
		LL_StepRssLimitSoft64
		LL_StepStackLimitHard64
		LL_StepStackLimitSoft64
		LL_StepWallClockLimitHard64
		LL_StepWallClockLimitSoft64
		LL_StepStepUserTime64
		LL_StepStepSystemTime64
		LL_StepStepMaxrss64
		LL_StepStepIxrss64
		LL_StepStepIdrss64
		LL_StepStepIsrss64
		LL_StepStepMinflt64
		LL_StepStepMajflt64
		LL_StepStepNswap64
		LL_StepStepInblock64
		LL_StepStepOublock64
		LL_StepStepMsgsnd64
		LL_StepStepMsgrcv64
		LL_StepStepNsignals64
		LL_StepStepNvcsw64
		LL_StepStepNivcsw64
		LL_StepStarterUserTime64
		LL_StepStarterSystemTime64
		LL_StepStarterMaxrss64
		LL_StepStarterIxrss64
		LL_StepStarterIdrss64
		LL_StepStarterIsrss64
		LL_StepStarterMinflt64
		LL_StepStarterMajflt64
		LL_StepStarterNswap64
		LL_StepStarterInblock64
		LL_StepStarterOublock64
		LL_StepStarterMsgsnd64
		LL_StepStarterMsgrcv64
		LL_StepStarterNsignals64
		LL_StepStarterNvcsw64
		LL_StepStarterNivcsw64
		LL_StepMachUsageCount
		LL_StepGetFirstMachUsage
		LL_StepGetNextMachUsage
		LL_StepCheckpointable
		LL_StepCheckpointing
		LL_StepCkptAccumTime
		LL_StepCkptFailStartTime
		LL_StepCkptFile
		LL_StepCkptGoodElapseTime
		LL_StepCkptGoodStartTime
		LL_StepCkptTimeHardLimit
		LL_StepCkptTimeHardLimit64
		LL_StepCkptTimeSoftLimit
		LL_StepCkptTimeSoftLimit64
		LL_StepCkptRestart
		LL_StepCkptRestartSameNodes
		LL_StepWallClockUsed
		LL_StepLargePage
		LL_StepMaxProtocolInstance
		LL_MachineAdapterList=800
		LL_MachineArchitecture
		LL_MachineAvailableClassList
		LL_MachineCPUs
		LL_MachineDisk
		LL_MachineFeatureList
		LL_MachineConfiguredClassList
		LL_MachineKbddIdle
		LL_MachineLoadAverage
		LL_MachineMachineMode
		LL_MachineMaxTasks
		LL_MachineName
		LL_MachineOperatingSystem
		LL_MachinePoolList
		LL_MachineRealMemory
		LL_MachineScheddRunningJobs
		LL_MachineScheddState
		LL_MachineScheddTotalJobs
		LL_MachineSpeed
		LL_MachineStartdState
		LL_MachineStartdRunningJobs
		LL_MachineStepList
		LL_MachineTimeStamp
		LL_MachineVirtualMemory
		LL_MachinePoolListSize
		LL_MachineFreeRealMemory
		LL_MachinePagesScanned
		LL_MachinePagesFreed
		LL_MachinePagesPagedIn
		LL_MachinePagesPagedOut
		LL_MachineGetFirstResource
		LL_MachineGetNextResource
		LL_MachineGetFirstAdapter
		LL_MachineGetNextAdapter
		LL_MachineDrainingClassList
		LL_MachineDrainClassList
		LL_MachineStartExpr
		LL_MachineSuspendExpr
		LL_MachineContinueExpr
		LL_MachineVacateExpr
		LL_MachineKillExpr
		LL_MachineDisk64
		LL_MachineRealMemory64
		LL_MachineVirtualMemory64
		LL_MachineFreeRealMemory64
		LL_MachinePagesScanned64
		LL_MachinePagesFreed64
		LL_MachinePagesPagedIn64
		LL_MachinePagesPagedOut64
		LL_MachineLargePageSize64
		LL_MachineLargePageCount64
		LL_MachineLargePageFree64
		LL_NodeTaskCount=1000
		LL_NodeGetFirstTask
		LL_NodeGetNextTask
		LL_NodeMaxInstances
		LL_NodeMinInstances
		LL_NodeRequirements
		LL_NodeInitiatorCount
		LL_SwitchTableJobKey=1200
		LL_TaskTaskInstanceCount=1400
		LL_TaskGetFirstTaskInstance
		LL_TaskGetNextTaskInstance
		LL_TaskExecutable
		LL_TaskExecutableArguments
		LL_TaskIsMaster
		LL_TaskGetFirstResourceRequirement
		LL_TaskGetNextResourceRequirement
		LL_TaskInstanceAdapterCount=1600
		LL_TaskInstanceGetFirstAdapter
		LL_TaskInstanceGetNextAdapter
		LL_TaskInstanceGetFirstAdapterUsage
		LL_TaskInstanceGetNextAdapterUsage
		LL_TaskInstanceMachineName
		LL_TaskInstanceTaskID
		LL_TaskInstanceMachineAddress
		LL_AdapterInterfaceAddress=1800
		LL_AdapterMode
		LL_AdapterName
		LL_AdapterUsageWindow
		LL_AdapterUsageProtocol
		LL_AdapterCommInterface=1806
		LL_AdapterUsageMode
		LL_AdapterTotalWindowCount=1811
		LL_AdapterAvailWindowCount
		LL_AdapterUsageAddress
		LL_AdapterUsageCommunicationInterface
		LL_AdapterUsageDevice
		LL_AdapterUsageInstanceNumber
		LL_AdapterWindowList
		LL_AdapterUsageWindowMemory64
		LL_AdapterMinWindowSize64
		LL_AdapterMaxWindowSize64
		LL_AdapterMemory64
		LL_CredentialGid=2000
		LL_CredentialGroupName
		LL_CredentialUid
		LL_CredentialUserName
		LL_StartdPerfJobsRunning=2200
		LL_StartdPerfJobsPending
		LL_StartdPerfJobsSuspended
		LL_StartdPerfCurrentJobs
		LL_StartdPerfTotalJobsReceived
		LL_StartdPerfTotalJobsCompleted
		LL_StartdPerfTotalJobsRemoved
		LL_StartdPerfTotalJobsVacated
		LL_StartdPerfTotalJobsRejected
		LL_StartdPerfTotalJobsSuspended
		LL_StartdPerfTotalConnections
		LL_StartdPerfFailedConnections
		LL_StartdPerfTotalOutTransactions
		LL_StartdPerfFailedOutTransactions
		LL_StartdPerfTotalInTransactions
		LL_StartdPerfFailedInTransactions
		LL_ScheddPerfJobsIdle=2400
		LL_ScheddPerfJobsPending
		LL_ScheddPerfJobsStarting
		LL_ScheddPerfJobsRunning
		LL_ScheddPerfCurrentJobs
		LL_ScheddPerfTotalJobsSubmitted
		LL_ScheddPerfTotalJobsCompleted
		LL_ScheddPerfTotalJobsRemoved
		LL_ScheddPerfTotalJobsVacated
		LL_ScheddPerfTotalJobsRejected
		LL_ScheddPerfTotalConnections
		LL_ScheddPerfFailedConnections
		LL_ScheddPerfTotalOutTransactions
		LL_ScheddPerfFailedOutTransactions
		LL_ScheddPerfTotalInTransactions
		LL_ScheddPerfFailedInTransactions
		LL_VersionCheck=2600
		LL_AdapterReqCommLevel=2700
		LL_AdapterReqUsage
		LL_ClusterGetFirstResource=2800
		LL_ClusterGetNextResource
		LL_ClusterSchedulingResources
		LL_ClusterDefinedResources
		LL_ClusterSchedulingResourceCount
		LL_ClusterDefinedResourceCount
		LL_ClusterEnforcedResources
		LL_ClusterEnforcedResourceCount
		LL_ClusterEnforceSubmission
		LL_ClusterSchedulerType
		LL_ResourceName=2900
		LL_ResourceInitialValue
		LL_ResourceAvailableValue
		LL_ResourceInitialValue64
		LL_ResourceAvailableValue64
		LL_ResourceRequirementName=3000
		LL_ResourceRequirementValue
		LL_ResourceRequirementValue64
		LL_WlmStatCpuTotalUsage=3100
		LL_WlmStatCpuSnapshotUsage
		LL_WlmStatMemoryHighWater
		LL_WlmStatMemorySnapshotUsage
		LL_MatrixTimeSlice = 3200
		LL_MatrixColumnCount
		LL_MatrixRowCount
		LL_MatrixGetFirstColumn
		LL_MatrixGetNextColumn
		LL_ColumnMachineName = 3300
		LL_ColumnProcessorNumber
		LL_ColumnRowCount
		LL_ColumnStepNames
		LL_MachUsageMachineName = 3400
		LL_MachUsageMachineSpeed
		LL_MachUsageDispUsageCount
		LL_MachUsageGetFirstDispUsage
		LL_MachUsageGetNextDispUsage
		LL_DispUsageEventUsageCount=3500
		LL_DispUsageGetFirstEventUsage
		LL_DispUsageGetNextEventUsage
		LL_DispUsageStepUserTime64
		LL_DispUsageStepSystemTime64
		LL_DispUsageStepMaxrss64
		LL_DispUsageStepIxrss64
		LL_DispUsageStepIdrss64
		LL_DispUsageStepIsrss64
		LL_DispUsageStepMinflt64
		LL_DispUsageStepMajflt64
		LL_DispUsageStepNswap64
		LL_DispUsageStepInblock64
		LL_DispUsageStepOublock64
		LL_DispUsageStepMsgsnd64
		LL_DispUsageStepMsgrcv64
		LL_DispUsageStepNsignals64
		LL_DispUsageStepNvcsw64
		LL_DispUsageStepNivcsw64
		LL_DispUsageStarterUserTime64
		LL_DispUsageStarterSystemTime64
		LL_DispUsageStarterMaxrss64
		LL_DispUsageStarterIxrss64
		LL_DispUsageStarterIdrss64
		LL_DispUsageStarterIsrss64
		LL_DispUsageStarterMinflt64
		LL_DispUsageStarterMajflt64
		LL_DispUsageStarterNswap64
		LL_DispUsageStarterInblock64
		LL_DispUsageStarterOublock64
		LL_DispUsageStarterMsgsnd64
		LL_DispUsageStarterMsgrcv64
		LL_DispUsageStarterNsignals64
		LL_DispUsageStarterNvcsw64
		LL_DispUsageStarterNivcsw64
		LL_EventUsageEventID = 3600
		LL_EventUsageEventName
		LL_EventUsageEventTimestamp
		LL_EventUsageStepUserTime64
		LL_EventUsageStepSystemTime64
		LL_EventUsageStepMaxrss64
		LL_EventUsageStepIxrss64
		LL_EventUsageStepIdrss64
		LL_EventUsageStepIsrss64
		LL_EventUsageStepMinflt64
		LL_EventUsageStepMajflt64
		LL_EventUsageStepNswap64
		LL_EventUsageStepInblock64
		LL_EventUsageStepOublock64
		LL_EventUsageStepMsgsnd64
		LL_EventUsageStepMsgrcv64
		LL_EventUsageStepNsignals64
		LL_EventUsageStepNvcsw64
		LL_EventUsageStepNivcsw64
		LL_EventUsageStarterUserTime64
		LL_EventUsageStarterSystemTime64
		LL_EventUsageStarterMaxrss64
		LL_EventUsageStarterIxrss64
		LL_EventUsageStarterIdrss64
		LL_EventUsageStarterIsrss64
		LL_EventUsageStarterMinflt64
		LL_EventUsageStarterMajflt64
		LL_EventUsageStarterNswap64
		LL_EventUsageStarterInblock64
		LL_EventUsageStarterOublock64
		LL_EventUsageStarterMsgsnd64
		LL_EventUsageStarterMsgrcv64
		LL_EventUsageStarterNsignals64
		LL_EventUsageStarterNvcsw64
		LL_EventUsageStarterNivcsw64
		LL_ClassName = 3700
		LL_ClassPriority
		LL_ClassExcludeUsers
		LL_ClassIncludeUsers = 3703
		LL_ClassExcludeGroups
		LL_ClassIncludeGroups
		LL_ClassAdmin
		LL_ClassNqsClass
		LL_ClassNqsSubmit
		LL_ClassNqsQuery
		LL_ClassMaxProcessors
		LL_ClassMaxJobs
		LL_ClassGetFirstResourceRequirement
		LL_ClassGetNextResourceRequirement
		LL_ClassComment
		LL_ClassCkptDir
		LL_ClassCkptTimeHardLimit
		LL_ClassCkptTimeSoftLimit
		LL_ClassWallClockLimitHard
		LL_ClassWallClockLimitSoft
		LL_ClassCpuStepLimitHard
		LL_ClassCpuStepLimitSoft
		LL_ClassCpuLimitHard
		LL_ClassCpuLimitSoft
		LL_ClassDataLimitHard
		LL_ClassDataLimitSoft
		LL_ClassCoreLimitHard
		LL_ClassCoreLimitSoft
		LL_ClassFileLimitHard
		LL_ClassFileLimitSoft
		LL_ClassStackLimitHard
		LL_ClassStackLimitSoft
		LL_ClassRssLimitHard
		LL_ClassRssLimitSoft
		LL_ClassNice
		LL_ClassFreeSlots
		LL_ClassMaximumSlots
		LL_ClassConstraints
		LL_ClassExecutionFactor
		LL_ClassMaxTotalTasks
		LL_ClassPreemptClass

	enum JobType:
		SET_BATCH
		SET_INTERACTIVE

	enum JobStepType:
		BATCH_JOB
		INTERACTIVE_JOB

	enum Usage:
		SHARED
		NOT_SHARED
		SLICE_NOT_SHARED

	enum LL_preempt_op:
		PREEMPT_STEP
		RESUME_STEP
		SYSTEM_PREEMPT_STEP

	enum LL_modify_op:
		EXECUTION_FACTOR
		CONSUMABLE_CPUS
		CONSUMABLE_MEMORY
		WCLIMIT_ADD_MIN
		JOB_CLASS
		ACCOUNT_NO

	ctypedef struct LL_modify_param:
		LL_modify_op type
		void *data

	enum LL_control_op:
		LL_CONTROL_RECYCLE
		LL_CONTROL_RECONFIG
		LL_CONTROL_START
		LL_CONTROL_STOP
		LL_CONTROL_DRAIN
		LL_CONTROL_DRAIN_STARTD
		LL_CONTROL_DRAIN_SCHEDD
		LL_CONTROL_PURGE_SCHEDD
		LL_CONTROL_FLUSH
		LL_CONTROL_SUSPEND
		LL_CONTROL_RESUME
		LL_CONTROL_RESUME_STARTD
		LL_CONTROL_RESUME_SCHEDD
		LL_CONTROL_FAVOR_JOB
		LL_CONTROL_UNFAVOR_JOB
		LL_CONTROL_FAVOR_USER
		LL_CONTROL_UNFAVOR_USER
		LL_CONTROL_HOLD_USER
		LL_CONTROL_HOLD_SYSTEM
		LL_CONTROL_HOLD_RELEASE
		LL_CONTROL_PRIO_ABS
		LL_CONTROL_PRIO_ADJ
		LL_CONTROL_START_DRAINED

	enum StepState:
		STATE_IDLE
		STATE_PENDING
		STATE_STARTING
		STATE_RUNNING
		STATE_COMPLETE_PENDING
		STATE_REJECT_PENDING
		STATE_REMOVE_PENDING
		STATE_VACATE_PENDING
		STATE_COMPLETED
		STATE_REJECTED
		STATE_REMOVED
		STATE_VACATED
		STATE_CANCELED
		STATE_NOTRUN
		STATE_TERMINATED
		STATE_UNEXPANDED
		STATE_SUBMISSION_ERR
		STATE_HOLD
		STATE_DEFERRED
		STATE_NOTQUEUED
		STATE_PREEMPTED
		STATE_PREEMPT_PENDING
		STATE_RESUME_PENDING

	enum SessionType:
		BATCH_SESSION
		INTERACTIVE_SESSION
		INTERACTIVE_HOSTLIST_SESSION
		
	enum CommLevel:
		LOW
		AVERAGE
		HIGH
	
	enum EventType:
		ERROR_EVENT=-1
		STATUS_EVENT
		TIMER_EVENT

	enum HoldType:
		NO_HOLD
		HOLDTYPE_USER
		HOLDTYPE_SYSTEM
		HOLDTYPE_USERSYS

	enum SpawnFlags:
		MARK_ALL_TASKS_RUNNING = (1<<0)

	ctypedef struct LL_STEP_ID:
		int  cluster
		int  proc
		char *from_host

	ctypedef struct LL_terminate_job_info:
		int version_num
		LL_STEP_ID StepId
		char *msg

	cdef struct ll_adapter_usage:
		char *dev_name
		char *protocol
		char *subsystem
		int  wid
		int  mem

	ctypedef ll_adapter_usage LL_ADAPTER_USAGE

	ctypedef struct LL_start_job_info_ext:
		int             version_num
		LL_STEP_ID      StepId
		char            **nodeList
		int             adapterUsageCount
		ll_adapter_usage *adapterUsage

	enum ckpt_type:
		CKPT_AND_CONTINUE
		CKPT_AND_TERMINATE
		CKPT_AND_HOLD

	enum wait_option_t:
		CKPT_NO_WAIT
		CKPT_WAIT

	enum CkptStart:
		CKPT_YES
		CKPT_NO
		CKPT_FAIL

	#
	# Data Access API functions
	#
	
	cdef extern char* c_version "ll_version" ()
	cdef extern LL_element *c_query "ll_query" (QueryType query_type)

	cdef extern int c_set_request "ll_set_request" (LL_element *,QueryFlags query_flag,char **,DataFilter)
	cdef extern int c_get_data "ll_get_data" (LL_element *,LLAPI_Specification, void *)
	cdef extern int c_reset_request "ll_reset_request" (LL_element *)
	cdef extern int c_deallocate "ll_deallocate" (LL_element *)
	cdef extern int c_free_objs "ll_free_objs" (LL_element *)
	cdef extern LL_element *c_get_objs "ll_get_objs" (LL_element *,LL_Daemon,char *,int *,int *)
	cdef extern LL_element *c_next_obj "ll_next_obj" (LL_element *)

	#
	# Error Handling API
	#

	cdef extern char *c_error "ll_error" (LL_element **, int)

	#
	# WorkLoad Management API
	#

	cdef extern int c_control "ll_control" (int,LL_control_op,char **,char **,char **,char **,int)
	cdef extern int c_modify "ll_modify" (int,LL_element *,LL_modify_param **,char **)
	cdef extern int c_preempt "ll_preempt" (int,LL_element **,char *,LL_preempt_op)
	cdef extern int c_terminate_job "ll_terminate_job" (LL_terminate_job_info *)
	cdef extern int c_start_job_ext "ll_start_job_ext" (LL_start_job_info_ext *ptr)

def ll_version():
	
	"""
	Returns a string of the  LoadLeveler version
	"""
	
	return c_version()

def ll_query(query_type):
	
	"""
	Defines the type of query to perform

	Parameters - query_type

		query_type - Query operation to perform

	Returns - A pointer to the LL_element object
	"""
	
	cdef LL_element *query_element
	cdef int QueryType
	
	QueryType = query_type 
	query_element = c_query(QueryType)
	
	retval = ptr_wrapper(query_element)
	
	return retval

def ll_set_request(query_element, query_flags, object_filter, data_filter):
	
	"""
	Determines the data requested by calls to getobjs

	Parameters - query_element, query_flags, object_filter, data_filter

		query_element - Pointer to the LL_element returned by ll_query

		query_flags - Query operation to perform

		object_filter - Search criteria for query_flags

		data_filter - Filter to apply to data returned by query_type

	Returns - Numeric, 0 for success or non-zero from ll_set_request routine
	"""
	
	cdef LL_element *QueryElement
	cdef char** ObjectFilter
	cdef int QueryFlags
	cdef int DataFilter
	
	QueryElement = ptr_unwrapper(query_element)
	QueryFlags = query_flags
	ObjectFilter = pyStringSeqToStringArray(object_filter)
	DataFilter = data_filter
	
	retval = c_set_request(QueryElement, QueryFlags, ObjectFilter, DataFilter)
	free(ObjectFilter)
	
	return retval

def ll_reset_request(query_element):

	"""
	Resets the query_element object to NULL so a new set_request
	can be used to request new data with the ll_get_objs routine.

	Parameters - query_element

		query_element - pointer to the LL_element returned by ll_query

	Returns - numeric, 0 for success and non-zero for error from ll_reset_request routine
	"""

	cdef LL_element *QueryElement
	cdef LL_element *cretval

	QueryElement = ptr_unwrapper(query_element)
	retval = c_reset_request(QueryElement)

	return retval

def ll_get_objs(query_element, ll_daemon, host_name = ""):
	
	"""
	Sends the query to the daemon requested

	Parameters - query_element, ll_daemon, host_name

		query_element - Pointer to the LL_element returned by ll_query

		ll_daemon - LoadLeveler daemon i.e LL_CM

		host_name - String, hostname where ll_daemon is located
 
	Returns - (retval, NumObjs, Errcode)

		retval - A pointer to the LL_element object

		NumObjs - Number of objects returned from ll_get_objs routine

		Errcode - 0 for success, other values that translate to a reason
	"""
	 
	cdef LL_element *QueryElement
	cdef int LL_Daemon
	cdef char *HostName
	cdef int NumObjs
	cdef int ErrCode
	cdef LL_element *cretval
	
	LL_Daemon = ll_daemon
	HostName = host_name
	QueryElement = ptr_unwrapper(query_element)
	
	cretval = c_get_objs(QueryElement, LL_Daemon, HostName, &NumObjs, &ErrCode)
	retval = ptr_wrapper(cretval)
	
	return retval, NumObjs, ErrCode
  
def ll_get_data(query_element, data_spec):
	
	"""
	Retrieves the data for the passed LL_element pointer

	Parameters - query_element, data_spec

		query_element - Pointer to the LL_element returned by ll_query

		data_spec - LLAPI specification .ie LL_MachineName

	Returns - Data which can be a string,number or list of strings/numbers or
		a pointer to the next LL_element object.
	"""
	
	cdef LL_element *QueryElement
	cdef int LL_DataSpec

	cdef int cdata
	cdef int LL_data_store_i
	cdef int* LL_data_store_ia
	cdef size_t LL_data_store_l
	cdef double LL_data_store_d
	cdef time_t LL_data_store_t
	cdef char* LL_data_store_c
	cdef char** LL_data_store_a
	cdef int  ipointer
	cdef char* pointer
	cdef LL_element *LL_data_store_e
	cdef LL_element *cretval
	
	QueryElement = ptr_unwrapper(query_element)
	LL_DataSpec = data_spec
	
	#
	# Query the C type for the requested LLAPI_Specification Enum
	#
	
	if pyloadlvar.LLAPI.has_key(data_spec):
		spec_string = pyloadlvar.LLAPI[data_spec]
		spec_type = pyloadlvar.LLTYPE[spec_string]
		#spec_string = pyloadlvar.LLAPI_ENUMS[data_spec]
		#print "LLAPI %d is type %d" % (data_spec,spec_type)
	else:
		spec_type = -1
		return "-1"

	# int * type handler
	
	if spec_type == 0:
		cdata = c_get_data(QueryElement, LL_DataSpec, &LL_data_store_i)
		retval = LL_data_store_i
	
	# int** type handler
	
	elif spec_type == 1:
		cdata = c_get_data(QueryElement, LL_DataSpec, &LL_data_store_ia)
		retval = []
		i = 0
		if data_spec == LL_MachinePoolList:
			cdata = c_get_data(QueryElement, LL_MachinePoolListSize, &LL_data_store_i)
			while i < LL_data_store_i:
				ipointer = LL_data_store_ia[i]
				retval.append(ipointer)
				i = i + 1
	
	# int64 type handler
	
	elif spec_type == 2:
		cdata = c_get_data(QueryElement, LL_DataSpec, &LL_data_store_l)
		retval = LL_data_store_l
	
	# time_t type handler
	
	elif spec_type == 3:
		cdata = c_get_data(QueryElement, LL_DataSpec, &LL_data_store_t)
		retval = LL_data_store_t
	
	# double type handler
	
	elif spec_type == 4:
		cdata = c_get_data(QueryElement, LL_DataSpec, &LL_data_store_d)
		retval = LL_data_store_d
	
	# char* type handler
	
	elif spec_type == 5:
		cdata = c_get_data(QueryElement, LL_DataSpec, &LL_data_store_c)
		retval = LL_data_store_c
	
	# char** type handler
	
	elif spec_type == 6:
		cdata = c_get_data(QueryElement, LL_DataSpec, &LL_data_store_a)
		retval = []
		i = 0
		pointer = LL_data_store_a[0]
		while pointer != NULL:
			retval.append(pointer)
			i = i + 1
			pointer = LL_data_store_a[i]
	
	# LL_element* type handler
	
	elif spec_type == 7:
		cdata = c_get_data(QueryElement, LL_DataSpec, &LL_data_store_e)
		retval = ptr_wrapper(LL_data_store_e)
	else:
		retval = ""

	return retval
	
def ll_next_obj(query_element):

	"""
	Obtains the next object in the query_element list specified

	Parameters - query_element

		query_element - Pointer to the LL_element returned by ll_query

	Returns - A wrapped pointer to the LL_element object
	"""
	
	cdef LL_element *QueryElement
	cdef LL_element *cretval
 
	QueryElement = ptr_unwrapper(query_element)

	cretval = c_next_obj(QueryElement)
	retval = ptr_wrapper(cretval)
 
	return retval

def ll_free_objs(query_element):
	
	"""
	Frees up the object returned from ll_get_objs routine

	Parameters - query_element

		query_element - pointer to the LL_element returned by ll_query

	Returns - Numeric reporting the return of the ll_free_objs routine
	"""
	
	cdef LL_element *QueryElement
	cdef int retval
	
	QueryElement = ptr_unwrapper(query_element)
	
	retval = c_free_objs(QueryElement)
	
	return retval

def ll_deallocate(query_element):
	
	"""
	Deallocates the query object

	Paramters - query_element

		query_element - Pointer to the LL_element returned by ll_query

	Returns - Numeric, reporting the return value of the ll_deallocate routine
	"""
	
	cdef LL_element *QueryElement
	cdef int retval
	
	QueryElement = ptr_unwrapper(query_element)
	
	retval = c_deallocate(QueryElement)
	
	return retval

def ll_error(error_obj, output_to):
	
	"""
	Return an error string equivalent to the error number

	Parameters - error_obj, output_to

		error_obj - Python C Object of the LoadLeveler error object

		output_to - Numeric, 1 prints to stdout, 2 prints to stderr
				
	Returns - Numeric, 0 for success or non-zero value if the operation failed 
	"""

	cdef LL_element *ErrorObj
	cdef int OutputTo
	cdef char* retval
	
	ErrorObj = ptr_unwrapper(error_obj)
	OutputTo = output_to
	
	retval = c_error(&ErrorObj, OutputTo)
	
	return retval

def ll_control(control, host_list=[], user_list=[], job_list=[], class_list=[], priority=0):

	"""
	Access wrapper functions llhold, llctl, llprio, llfavoruser, llfavorjob

	Parameters - control, host_list, user_list, job_list, class_list, priority

		control - Operation to perform

		host_list - List of hosts to perform operation against

		user_list - List of users to perform operation against

		job_list - List of jobs to perform operation against

		class_list - List of classes to perform operation against

		priority - Numeric value of the priority adjustment

	Returns - Numeric, the return value from the ll_control routine
	"""

	cdef int Version
	cdef int Control_op
	cdef char** Hosts
	cdef char** Users
	cdef char** Jobs
	cdef char** JClass
	cdef int Priority
	cdef int retval

	retval = 0
	Version = pyloadlvar.LL_CONTROL_VERSION
	Control_op = control
	Priority = priority

	Hosts  = pyStringSeqToStringArray(host_list)
	Users  = pyStringSeqToStringArray(user_list)
	Jobs   = pyStringSeqToStringArray(job_list)
	JClass = pyStringSeqToStringArray(class_list)

	retval = c_control(Version,Control_op,Hosts,Users,Jobs,JClass,Priority)

	free(Hosts)
	free(Users)
	free(Jobs)
	free(JClass)

	return retval

def llctl(control_op,class_list=[],host_list=[]):

	"""
	Main control function called by wrapper functions llfavorjob, llfavoruser, llprio & llhold

	Parameters - control_op, class_list, host_list

		control_op - Operation to perform i.e

			LL_CONTROL_START
			LL_CONTROL_STOP
			LL_CONTROL_RECYCLE
			LL_CONTROL_RECONFIG
			LL_CONTROL_DRAIN
			LL_CONTROL_DRAIN_SCHEDD
			LL_CONTROL_DRAIN_STARTD
			LL_CONTROL_FLUSH
			LL_CONTROL_PURGE_SCHEDD
			LL_CONTROL_SUSPEND
			LL_CONTROL_RESUME
			LL_CONTROL_RESUME_STARTD
			LL_CONTROL_RESUME_SCHEDD
			LL_CONTROL_START_DRAINED

		class_list - List of classes to perform operation against

		host_list - List of hosts to perform operation against

	Returns - Numeric, return value from control function 
	"""
 
	if control_op != LL_CONTROL_START and control_op != LL_CONTROL_STOP and \
	control_op != LL_CONTROL_RECYCLE and control_op != LL_CONTROL_RECONFIG and \
	control_op != LL_CONTROL_DRAIN and control_op != LL_CONTROL_DRAIN_SCHEDD and \
	control_op != LL_CONTROL_DRAIN_STARTD and control_op != LL_CONTROL_FLUSH and \
	control_op != LL_CONTROL_PURGE_SCHEDD and control_op != LL_CONTROL_SUSPEND and \
	control_op != LL_CONTROL_RESUME or control_op != LL_CONTROL_RESUME_STARTD and \
	control_op != LL_CONTROL_RESUME_SCHEDD and control_op != LL_CONTROL_START_DRAINED:
		return -1
	else:
		return ll_control(control_op,host_list,[],[],class_list,0)

def llfavorjob(control_op,job_list=[]):

	"""
	Wrapper function for llcontrol to favor a list of jobs

	Parameters - control_op, job_list

		control_op - Operation to perform i.e

			LL_CONTROL_FAVOR_JOB
			LL_CONTROL_UNFAVOR_JOB

		job_list - List of jobsteps to favor operation against

	Returns - Numeric, return value from control function 
	"""

	class_list = []
	host_list  = []
	user_list  = []

	if control_op != LL_CONTROL_FAVOR_JOB and control_op != LL_CONTROL_UNFAVOR_JOB:
		return -1
	else:
		return ll_control(control_op,host_list,user_list,job_list,class_list,0)

def llfavoruser(control_op,user_list=[]):

	"""
	Wrapper function for llcontrol to favor a list of users

	Parameters - control_op, user_list

		control_op - Operation to perform i.e

			LL_CONTROL_FAVOR_USER
			LL_CONTROL_UNFAVOR_USER

		user_list - List of users to favor

	Returns - Numeric, return value from control function 
	"""

	class_list = []
	host_list  = []
	job_list   = []

	if control_op != LL_CONTROL_FAVOR_USER and control_op != LL_CONTROL_UNFAVOR_USER:
		return -1
	else:
		return ll_control(control_op,host_list,user_list,job_list,class_list,0)

def llhold(control_op,host_list=[],user_list=[],job_list=[]):

	"""
	Wrapper function for llcontrol to hold a list of jobs or list of users

	Parameters - control_op, job_list

		control_op - Operation to perform

			LL_CONTROL_HOLD_USER
			LL_CONTROL_HOLD_SYSTEM
			LL_CONTROL_HOLD_RELEASE

		host_list  - List of hosts

		user_list - List of users
 
		job_list - List of jobsteps

	Returns - Numeric, return value from control function 
	"""

	class_list = []

	if control_op != LL_CONTROL_HOLD_USER and control_op != LL_CONTROL_HOLD_SYSTEM \
	and control_op != LL_CONTROL_HOLD_RELEASE:
		return -1
	else:
		return ll_control(control_op,host_list,user_list,job_list,class_list,0)

def llprio(control_op,job_list,priority):

	"""
	Wrapper function for llcontrol to alter the priority of a list of jobsteps

	Parameters - control_op, job_list, priority

		control_op - Priority operation to perform

				LL_CONTROL_PRIO_ABS - New absolute priority value
				LL_CONTROL_PRIO_ADJ - New adjusted priority value

		job_list - List of jobsteps

		priority - Numeric, the priority value 

	Returns - Numeric, return value from control function 
	"""

	class_list = []
	host_list  = []
	user_list  = []

	if control_op != LL_CONTROL_PRIO_ABS and control_op != LL_CONTROL_PRIO_ADJ:
		return -1
	else:
		return ll_control(control_op,host_list,user_list,job_list,class_list,priority)

def ll_modify(modify_op, modify_data, job_id):

	"""
	Modify the attributes of jobstep

	Parameters - modify_op, modify_data, job_id

		modify_op - Operation to perform i.e

			EXECUTION_FACTOR
			CONSUMABLE_CPUS
			CONSUMABLE_MEMORY
			WCLIMIT_ADD_MIN
			JOB_CLASS
			ACCOUNT_NO

		modify_data - Parameter to accompany modify operation

		job_id - String in the form host.job_id.step_id i.e shivling.5.0

	Returns - Numeric value indicating the return code of the operation,
	the second is a pointer to the ll_error object returning an error string.
	"""
 
	cdef int Version
	cdef LL_element *ErrorObj
	cdef int LL_modify_op
	cdef int LL_modify_data_i
	cdef char* LL_modify_data_c
	cdef LL_modify_param mycmd, *cmdp[2]
	cdef char** Job_list
	cdef int retval

	Version = pyloadlvar.LL_API_VERSION
	ErrorObj = NULL

	LL_modify_op = modify_op
	mycmd.type = LL_modify_op

	if modify_op == EXECUTION_FACTOR or modify_op == CONSUMABLE_CPUS \
	or modify_op == WCLIMIT_ADD_MIN:
		LL_modify_data_i = modify_data
		mycmd.data = &LL_modify_data_i
	elif modify_op == CONSUMABLE_MEMORY:
		LL_modify_data_i = modify_data
		mycmd.data = &LL_modify_data_i
	elif modify_op == JOB_CLASS or modify_op == ACCOUNT_NO:
		LL_modify_data_c = modify_data
		mycmd.data = LL_modify_data_c

	cmdp[0] = &mycmd
	cmdp[1] = NULL

	Job_list = pyStringSeqToStringArray(job_id)

	retval = c_modify(Version,&ErrorObj,cmdp,Job_list)
	cretval = ptr_wrapper(ErrorObj)

	free(Job_list)

	return retval, cretval

def ll_preempt(job_step, preempt_op):

	"""
	Preempts a currently running jobstep or resumes a preempted jobstep within the GANG scheduler

	Parameters - job_step, preempt_op

		job_step - String in the form host.job_id.step_id i.e shivling.5.0

		preempt_op - Operation to perform i.e

			PREEMPT_STEP
			RESUME_STEP

	Returns - A numeric value indicating the return code of the operation
	and the second is a pointer to the ll_error object 
	"""

	cdef int Version
	cdef LL_element *ErrorObj
	cdef int LL_preempt_op
	cdef char* Job_Step
	cdef int retval

	Version = pyloadlvar.LL_API_VERSION
	ErrorObj = NULL
	Job_Step = job_step
	LL_preempt_op = preempt_op

	retval = c_preempt(Version,&ErrorObj,Job_Step,LL_preempt_op)

	cretval = ptr_wrapper(ErrorObj)

	return retval, cretval

def ll_terminate_job(cluster, proc, host, msg=""):

	"""
	Instructs the negotiator deamon to cancel the specified jobstep

	Parameters - cluster, proc, host, msg

		cluster - Numeric, the jobid

		proc - Numeric, the stepid

		host - String, the submitting host

		msg - String, message available via get_data subroutine as to why the job was cancelled

	Returns - Numeric, 0 for success else a non-zero representing the error condition
	"""

	cdef int Version
	cdef int Cluster
	cdef int Proc
	cdef char* Job_Host
	cdef char* Job_Msg
	cdef LL_terminate_job_info *cancel_info
	cdef int retval

	Version = pyloadlvar.LL_PROC_VERSION
	Cluster = cluster
	Proc = proc
	Job_Host = host
	Job_Msg = msg

	cancel_info = <LL_terminate_job_info *> malloc(sizeof(LL_terminate_job_info))
	cancel_info.version_num = Version
	cancel_info.msg = Job_Msg
	cancel_info.StepId.cluster= Cluster
	cancel_info.StepId.proc = Proc
	cancel_info.StepId.from_host = Job_Host

	retval = c_terminate_job(cancel_info)
	free(cancel_info)

	return retval

def ll_start_job_ext(step_id,node_list,adapter_list):

	"""
	Start a jobstep via an external scheduler with the SCHEDULER=API settings in the LoadL_config file

	Parameters - step_id, node_list, adapter_list

		step_id - String in the form host.step_id.step_no i.e shivling.5.0 
                 
		node_list - List should contain a node entry for each task i.e

			[ 'nodeA', 'nodeA', 'nodeB', 'nodeB' ]

		adapter_list - List containing adapter info for each task matching each entry in node_list

			[ [dev_name, protocol, subsystem, win_id, memory] ... ]

			dev_name is the device name - en0, css0 etc
			protocol is either MPI or LAPI
			subsystem is either US or IP
			win_id is the window ID on the switch adapter
			memory is the memory to be used on the switch adapter

			NB -
				if a non switch adapter is used then both wid & mem should be set to 0
				if a hybrid MPI/LAPI task is required then

	Returns - Numeric, 0 for success else a non-zero value representing an error condition
        """

	cdef int Version
	cdef char* Host
	cdef char** Hosts
	cdef int Cluster
	cdef int Proc
	cdef int retval
	cdef int Adapter_Num
	cdef LL_start_job_info_ext *start_info

	Adapter_Num = len(node_list)
	Hosts = pyStringSeqToStringArray(node_list)

	Version = pyloadlvar.LL_PROC_VERSION
	(host,cluster,proc) = string.split(step_id,'.')

	Cluster = int(cluster)
	Proc = int(proc)
	Host = host

	# Populate job_info structure 

	start_info.StepId.cluster = Cluster
	start_info.StepId.proc = Proc
	start_info.StepId.from_host = Host
	start_info.version_num = Version
	start_info.nodeList = Hosts
	start_info.adapterUsageCount = Adapter_Num

	start_info.adapterUsage = <LL_ADAPTER_USAGE *> malloc(start_info.adapterUsageCount * sizeof(LL_ADAPTER_USAGE))

	# Populate adapter_list structure with adapter information

	node_num = 0
	for node in node_list:

		dev_name = adapter_list[node_num][0]
		protocol = adapter_list[node_num][1]
		subsystem = adapter_list[node_num][2]
		wid = adapter_list[node_num][3]
		mem = adapter_list[node_num][4]

		start_info.adapterUsage[node_num].dev_name = dev_name
		start_info.adapterUsage[node_num].protocol = protocol
		start_info.adapterUsage[node_num].subsystem = subsystem
		start_info.adapterUsage[node_num].wid = wid
		start_info.adapterUsage[node_num].mem = mem

		node_num = node_num + 1

	retval = c_start_job_ext(start_info)

	free(start_info)
	
	return retval

def PyCObjValid(query_element):

	"""
	Wrapper function to check if the returned LL_element object is NULL

	Parameter - query_element

		query_element - Pointer to the LL_element returned by ll_query

	Returns - Numeric, 0 if no data exists, 1 if data exists
	"""
	
	cdef LL_element *QueryElement
	cdef int retval

	retval = 1	
	QueryElement = ptr_unwrapper(query_element)

	if QueryElement == NULL:
		retval = 0
		
	return retval
