"""PyLoadL: Python interface to IBM LoadLeveler

PyLoadL is a Pyrex wrapper that provides an interface 
to the IBM LoadLeveler batch scheduler API
"""

from distutils.core import setup
from distutils.extension import Extension
from distutils.sysconfig import get_python_lib
from Pyrex.Distutils import build_ext

import os
import sys, popen2
from string import *

#
# Python 2.3.3 version required because that's what I've tested
#

if not (sys.version_info[0] >= 2 and sys.version_info[1] >= 3 and sys.version_info[2] >= 3):
    print "I suggest using Python 2.3.3 or greater to build PyLoadL. Exiting..."
    sys.exit(1)

#
# Get the LoadLeveler version from AIX lslpp package utility
#

pout,pin,perr= popen2.popen3('/usr/bin/lslpp -Ou -qlc LoadL.full',)
pin.close()
a = pout.readline()
pout.close()
b = perr.readline()
perr.close()

#
# Create a version string similar to the LoadLeveler PERL module alla Mike Hawkins
# for creating LoadLeveler version dependant pyloadl code in the near future
#

LLVER = ''
if ( not b ):
	c = a.split(':')
        ver_str = c[2].split('.')
        LLVER = LLVER + ver_str[0]
	for a in ver_str[1:]:
		if len(a) < 2:
			a = '0' + a
		LLVER = LLVER + a
else:
	print "lslpp error detected querying LoadLeveler installp package\n"
	print "%s" % b
	print "Exiting ...."
	sys.exit(1)

compiler_dir = os.path.join(get_python_lib(prefix=''), 'pyloadl/')

#
# Standard LoadLeveler location under AIX
#

include_dirs = ['/usr/lpp/LoadL/full/include','/usr/include']
library_dirs = ['/usr/lpp/LoadL/full/lib','/usr/lib']
libraries = ['llapi']

#
# Trove classifiers
#

classifiers = """\
Development Status :: 3 - Alpha
Intended Audience :: Developers
License :: OSI Approved :: GNU General Public License (GPL)
Natural Language :: English
Operating System :: POSIX :: AIX
Programming Language :: Python
Topic :: Software Development :: Libraries :: Python Modules
"""
doclines = __doc__.split("\n")

setup(
    name = "PyLoadL",
    version = "0.0.1a4",
    description = doclines[0],
    long_description = "\n".join(doclines[2:]),
    author = "Mark Roberts",
    author_email = "mark at gingergeeks co uk",
    url = "http://www.gingergeeks.co.uk/pyloadl/",
    classifiers = filter(None, classifiers.split("\n")),
    platforms = ["AIX"],
    keywords = ["IBM", "LoadLeveler"],
    packages = ["pyloadl"],
    ext_modules = [
        Extension( "pyloadl",
                    ["pyloadl/pyloadl.pyx"],
                    libraries = libraries,
                    library_dirs = library_dirs,
                    include_dirs = include_dirs)
    ],
    cmdclass = {"build_ext": build_ext}
)
